/*
 * @file  EnumerationDefs.h
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Mar 2, 2012
 * @version 1.0
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 *	Template definitions for the Enumeration class
 */

#ifndef ENUMERATIONDEFS_H_
#define ENUMERATIONDEFS_H_

template<unsigned N>
array<vector<pair<unsigned char, double>> , N / 2> split(const array<vector<pair<unsigned char, double>> , N>& pDistribution,
		bool pLeft) {
	array<vector<pair<unsigned char, double>> , N / 2> splitDistribution;
	auto left = pDistribution.begin() + (pLeft ? 0 : N / 2);
	std::copy(left, left + N / 2, splitDistribution.begin());
	return splitDistribution;
}

template<unsigned N>
array<vector<pair<unsigned char, double>> , N> indexDistribution(const array<vector<double> , N>& pDistribution) {
	array<vector<pair<unsigned char, double>> , N> distribution;
	for(uint64_t i0 = 0; i0 < N; ++i0)
	for(uint64_t i1 = 0; i1 < 0x100; ++i1)
	distribution[i0].push_back( {(unsigned char)i1, pDistribution[i0][i1]});
	return distribution;
}

template<unsigned N>
Enumeration<N>::Enumeration(const array<vector<double> , N>& pDistribution) :
			//	Enumeration(expandIndex<N>(pDistribution)) { // c++11 not fully supported yet
			mEnumerationX(split<N> (indexDistribution<N> (pDistribution), true)),
			mEnumerationY(
					split<N> (indexDistribution<N> (pDistribution), false)) {
	addCandidate(0, 0);
}

template<unsigned N>
Enumeration<N>::Enumeration(const array<vector<pair<unsigned char, double>> , N>& pDistribution) :
mEnumerationX(split<N>(pDistribution, true)), mEnumerationY(split<N> (pDistribution,
				false)) {
	addCandidate(0, 0);
}

template<unsigned N>
Enumeration<N>::~Enumeration() {
}

template<unsigned N>
std::pair<double, array<unsigned char, N>> Enumeration<N>::nextCandidate() {
	if (mCandidates.isEmpty())
	throw std::runtime_error("Empty Candidate Set");

	auto next(extractCandidate());

	addCandidate(next.getIndexX() + 1, next.getIndexY());
	addCandidate(next.getIndexX(), next.getIndexY() + 1);

	return next.makePair();
}

template<unsigned N>
Event<N> Enumeration<N>::extractCandidate() {
	Event<N> extracted = std::move(mCandidates.pop());
	mListX.clearMark(extracted.getIndexX());
	mListY.clearMark(extracted.getIndexY());
	return extracted;
}

template<unsigned N>
uint64_t Enumeration<N>::memSize() const {
	uint64_t sizeValue = 0;
	sizeValue += mEnumerationX.memSize();
	sizeValue += mEnumerationY.memSize();
	sizeValue += mCandidates.memSize();
	sizeValue += mListX.memSize();
	sizeValue += mListY.memSize();
	return sizeValue;
}

template<unsigned N>
uint64_t Enumeration<N>::diskSize() const {
	uint64_t sizeValue = 0;
	sizeValue += mEnumerationX.diskSize();
	sizeValue += mEnumerationY.diskSize();
	sizeValue += mListX.diskSize();
	sizeValue += mListY.diskSize();
	return sizeValue;
}

template<unsigned N>
bool Enumeration<N>::addCandidate(uint64_t pX, uint64_t pY) {
	if (N < 0x10) {
		//if ((pX | pY) >> (N << 2)) // g++ throws a warning on this...
			if (((pX | pY) >> (N << 1)) >> (N << 1))
			return false;
	}
	if (pX >= mListX.size())
		mListX.insert(mEnumerationX.nextCandidate());
	if (mListX.isMarked(pX))
		return false;

	if (pY >= mListY.size())
		mListY.insert(mEnumerationY.nextCandidate());
	if (mListY.isMarked(pY))
		return false;

	mCandidates.push(mListX.getEvent(pX), mListY.getEvent(pY));
	mListX.setMark(pX);
	mListY.setMark(pY);
	return true;
}

#endif /* ENUMERATIONDEFS_H_ */
