/*
 * @file  Event.h
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Mar 2, 2012
 * @version 1.0
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 * Placeholder class for key candidates
 */

#ifndef EVENT_H_
#define EVENT_H_

#include "Includes.h"

template<unsigned N>
class Event {
public:
	Event() {
	}

	Event(uint64_t pIndex, const pair<double, array<unsigned char, N>>& pEvent) :
	mProbability(pEvent.first), mIndexX(pIndex), mLabel(pEvent.second) {
	}

	Event(const Event<N / 2>& pEvent1, const Event<(N + 1) / 2>& pEvent2) :
		mProbability(pEvent1.getProbability() * pEvent2.getProbability()), mIndexX(
				pEvent1.getIndexX()), mIndexY(pEvent2.getIndexX()) {
		auto label1 = pEvent1.getLabel();
		copy(label1.begin(), label1.end(), mLabel.begin());
		auto label2 = pEvent2.getLabel();
		copy(label2.begin(), label2.end(), mLabel.begin() + N / 2);
	}

	virtual ~Event() {
	}

	inline double getProbability() const {
		return mProbability;
	}
	inline uint64_t getIndexX() const {
		return mIndexX;
	}
	inline uint64_t getIndexY() const {
		return mIndexY;
	}

	inline  array<unsigned char, N> getLabel() const {
		return mLabel;

	}

	inline pair<double, array<unsigned char, N>> makePair() const {
		return make_pair(mProbability, mLabel);
	}

	inline bool operator()(const Event& pEvent1, const Event& pEvent2) const {
		return pEvent1.getProbability() < pEvent2.getProbability();
	}

protected:
	double mProbability;
	uint64_t mIndexX;
	uint64_t mIndexY;
	array<unsigned char, N> mLabel;
};

#endif /* EVENT_H_ */
