/*
 * EventQueue.h
 *
 *  Created on: Mar 23, 2012
 *      Author: nicolaq
 */

#ifndef EVENTQUEUE_H_
#define EVENTQUEUE_H_

#include "Includes.h"
#include"Event.h"

#ifndef LAZYQUEUE_OPT
template<unsigned N>
class EventQueue {
public:
	EventQueue() {
	}
	virtual ~EventQueue() {
	}

	void push(const Event<N / 2>& pEvent1, const Event<(N + 1) / 2>& pEvent2) {
		mQueue.emplace(pEvent1, pEvent2);
	}
	Event<N> pop() {
		Event<N> extracted = mQueue.top();
		mQueue.pop();
		return extracted;
	}

	bool isEmpty() const {
		return mQueue.empty();
	}

	uint64_t memSize() const {
		return sizeof(*this) + mQueue.size() * sizeof(Event<N>);
	}
protected:
priority_queue<Event<N>, vector<Event<N>>, Event<N>> mQueue;
};
#else
template<unsigned N>
class EventQueue {
public:
	EventQueue() :
		mReduced(true), mQueue(1) {
	}
	virtual ~EventQueue() {
	}

	void push(const Event<N / 2>& pEvent1, const Event<(N + 1) / 2>& pEvent2) {
		if (mReduced) {
			uint64_t index = mQueue.size();
			mQueue.emplace_back(pEvent1, pEvent2);
			while (index > 1 && mQueue[index >> 1].getProbability()
					< mQueue[index].getProbability()) {
				swap(mQueue[index >> 1], mQueue[index]);
				index >>= 1;
			}
		} else {
			mQueue[1] = move(Event<N>(pEvent1, pEvent2));
			maxHeap(1);
			mReduced = true;
		}
		++mSize;
	}

	Event<N> pop() {
		if(!mReduced) {
			mQueue[1] = move(mQueue.back());
			mQueue.pop_back();
			maxHeap(1);
		}
		Event<N> extracted = move(mQueue[1]);
		mReduced = false;
		mSize--;
		return extracted;
	}

	bool isEmpty() const {
		return mSize == 0;
	}

	uint64_t memSize() const {
		return sizeof(*this) + mQueue.size() * sizeof(Event<N>);
	}
protected:
	void maxHeap(uint64_t index) {
		uint64_t largest = index;
		uint64_t child = index << 1;
		if (child < mQueue.size() && mQueue[largest].getProbability()
				< mQueue[child].getProbability())
			largest = child;
		++child;
		if (child < mQueue.size() && mQueue[largest].getProbability()
				< mQueue[child].getProbability())
			largest = child;
		if (largest != index) {
			swap(mQueue[largest], mQueue[index]);
			maxHeap(largest);
		}
	}

	uint64_t mSize;
	bool mReduced;
	vector<Event<N>> mQueue;
};
#endif
#endif /* EVENTQUEUE_H_ */
