/*
 * ProcessedEnumeration.cpp
 *
 *  Created on: Apr 4, 2012
 *      Author: nicolaq
 */

#include"ProcessedEnumeration.h"

ProcessedEnumeration::ProcessedEnumeration(
		const array<vector<double> , 0x10>& pDistribution) :
	mEnumeration(process(pDistribution)) {
}

ProcessedEnumeration::~ProcessedEnumeration() {
}

pair<double, array<unsigned char, 0x10>> ProcessedEnumeration::nextCandidate() {
	return process(mEnumeration.nextCandidate());
}

uint64_t ProcessedEnumeration::memSize() const {
	return mEnumeration.memSize();
}

array<vector<double> , 0x10> ProcessedEnumeration::process(const array<vector<
		double> , 0x10>& pDistribution) {
	map<double, vector<uint64_t>> group;
	for (unsigned i0 = 0; i0 < 0x10; ++i0) {
double	entropy = std ::accumulate(pDistribution[i0].begin(), pDistribution[i0].end(), 0.0,
			[] (double acc, double proba) {return acc - proba * log(proba);
			});
	group[entropy] = {i0};

};
while(group.size() > 1) {
	map<double, vector<uint64_t>> newGroup;
	auto it1 = group.begin();
	auto it2 = group.rbegin();
	for(unsigned i0 = 0; i0 < (group.size() >> 1); ++i0) {
		vector<uint64_t> concatenate(it1->second);
		concatenate.insert(concatenate.end(), it2->second.begin(), it2->second.end());
		newGroup[it1->first + it2->first] = concatenate;
		++it1;
		++it2;
	}
	group = newGroup;
}
array<vector<double> , 0x10> shuffledDistribution;
for(unsigned i0 = 0; i0 < 0x10; ++i0) {
	mShuffle[i0] = group.begin()->second[i0];
	shuffledDistribution[i0] = pDistribution[mShuffle[i0]];
}
return shuffledDistribution;
}

pair<double, array<unsigned char, 0x10>> ProcessedEnumeration::process(const pair<double, array<unsigned char, 0x10>>& pCandidate) {
	pair<double, array<unsigned char, 0x10>> candidate;
	candidate.first = pCandidate.first;
	for(unsigned i0 = 0; i0 < 0x10; ++i0)
	candidate.second[mShuffle[i0]] = pCandidate.second[i0];
	return candidate;
}
