/*
 * SplitDistribution.cpp
 *
 *  Created on: Apr 16, 2012
 *      Author: nicolaq
 */

#include "SplitDistribution.h"

pair<double, uint64_t> splitEntropy(const DistributionType& pDistribution) {
	double entropy = 0;
	uint64_t index = 0, maxIndex = 0;
	for(auto& v0 : pDistribution) {
		double subkeyEntropy = accumulate(v0.begin(), v0.end(), 0.0,
				[] (double acc, pair<unsigned char, double> proba) {return acc - proba.second * log(proba.second);
				});
		if(subkeyEntropy > entropy) {
			maxIndex = index;
			entropy = subkeyEntropy;
		}
		index++;
	}
	return make_pair(entropy, maxIndex);
}

pair<double, DistributionType> split(const DistributionType& pDistribution,
		bool pLeft) {
	uint64_t splitIndex = splitEntropy(pDistribution).second;
	DistributionType distribution;
	for (unsigned i0 = 0; i0 < pDistribution.size(); ++i0)
		if (i0 != splitIndex)
			distribution[i0] = pDistribution[i0];
		else {
			auto subDistribution = pDistribution[i0];
			sort(subDistribution.begin(), subDistribution.end(),
					[subDistribution](pair<unsigned char, double> v1, pair<unsigned char, double> v2) {return v1.second > v2.second;});
			for (unsigned i1 = 0; i1 < subDistribution.size(); ++i1)
				if ((i1 & 0x1) ^ pLeft)
					distribution[i0].push_back(subDistribution[i1]);
		}
return {	-splitEntropy(distribution).first, distribution};
}

multimap<double, DistributionType> splitDistribution(
		const DistributionType& pDistribution, uint64_t numSplit) {
	multimap<double, DistributionType> splitMap;
	splitMap.insert( {-splitEntropy(pDistribution).first, pDistribution});
	for (uint64_t i0 = 1; i0 < numSplit; ++i0) {
		auto nextSplit = splitMap.begin()->second;
		splitMap.erase(splitMap.begin());
		splitMap.insert(split(nextSplit, true));
		splitMap.insert(split(nextSplit, false));
	}
	return splitMap;
}

array<vector<pair<unsigned char, double>>, 0x10> indexDistribution(const array<vector<double> , 0x10>& pDistribution) {
	array<vector<pair<unsigned char, double>> , 0x10> distribution;
	for(uint64_t i0 = 0; i0 < 0x10; ++i0)
	for(uint64_t i1 = 0; i1 < 0x100; ++i1)
	distribution[i0].push_back( {(unsigned char)i1, pDistribution[i0][i1]});
	return distribution;
}
