/*
 * ThreadEnumeration.cpp
 *
 *  Created on: Apr 10, 2012
 *      Author: nicolaq
 */

#include<iostream>

#include "ThreadEnumeration.h"
#include "SplitDistribution.h"

ThreadEnumeration::ThreadEnumeration(
		const array<vector<double> , 0x10>& pDistribution,
		uint64_t pNumThreads, uint64_t pNumSplit, uint64_t pBufferSize) :
	mRun(true), mBufferSize(pBufferSize), mBufferMask(pBufferSize - 1),
			mBufferHead(0), mBufferTail(0), mBuffer(pBufferSize), mHeap(1),
			mSplitBufferHead(pNumSplit, 0), mSplitBufferTail(pNumSplit, 0),
			mSplitBuffer(pNumSplit, vector<Candidate> (mBufferSize)),
			mEnumLocks(pNumSplit), mThreadDelay(100) {
	auto splitMap = splitDistribution(indexDistribution(pDistribution), pNumSplit);
	for(auto& v0 : splitMap)
	mSplitEnums.push_back(unique_ptr<Enumeration<0x10>>(new Enumeration<0x10>(v0.second)));

	unique_lock<mutex> heapLock(mHeapLock);
	// create threads and put them to work
	for (unsigned i0 = 0; i0 < pNumThreads; ++i0)
	mThreads.push_back(thread {&ThreadEnumeration::doThread, this});
	// initialize heap
	while (*min_element(mSplitBufferTail.begin(), mSplitBufferTail.end()) == 0)
		this_thread::sleep_for(mThreadDelay);
	for (unsigned i0 = 0; i0 < mSplitBufferHead.size(); ++i0) {
		pushHeap(i0, mSplitBuffer[i0][mSplitBufferHead[i0] & mBufferMask]);
		++mSplitBufferHead[i0];
	}
}

ThreadEnumeration::~ThreadEnumeration() {
	// tell threads to die and wait for them to do it
	mRun = false;
for(auto& v0: mThreads)
v0.join();
}

Candidate ThreadEnumeration::nextCandidate() {
	while (true) {
		if (mBufferHead != mBufferTail) {
			auto extracted = mBuffer[mBufferHead & mBufferMask];
			++mBufferHead;
			return extracted;
		}
		if (!fillSplitBuffer() && !fillBuffer()) {
			this_thread::sleep_for(mThreadDelay);
		}
	}
}

uint64_t ThreadEnumeration::memSize() const {
	uint64_t sizeValue = 0;
	for(auto& v0 : mSplitEnums)
	sizeValue += v0->memSize();
	sizeValue += sizeof(*this);
	sizeValue += mSplitBuffer.size() * sizeof(atomic<uint64_t> );
	sizeValue += mSplitBuffer.size() * mBufferSize * sizeof(pair<double, array<unsigned char, 0x10>>);
	sizeValue += mEnumLocks.size() * sizeof(mutex);
	sizeValue += mSplitEnums.size() * sizeof(Enumeration<0x10>*);
	return sizeValue;
}

uint64_t ThreadEnumeration::diskSize() const {
	uint64_t sizeValue = 0;
	for(auto& v0 : mSplitEnums)
	sizeValue += v0->diskSize();
	return sizeValue;
}

void ThreadEnumeration::doThread() {
	while (mRun)
		if(!fillSplitBuffer() && !fillBuffer())
			this_thread::sleep_for(mThreadDelay);
}

bool ThreadEnumeration::fillSplitBuffer() {
	for (unsigned i0 = 0; i0 < mSplitEnums.size(); i0++)
		if (mSplitBufferHead[i0] + mBufferMask > mSplitBufferTail[i0]) {
			unique_lock<mutex> enumLock(mEnumLocks[i0], defer_lock_t());
			if (enumLock.try_lock()) {
				while (mSplitBufferHead[i0] + mBufferMask
						> mSplitBufferTail[i0]) {
					mSplitBuffer[i0][mSplitBufferTail[i0] & mBufferMask]
							= mSplitEnums[i0]->nextCandidate();
					++mSplitBufferTail[i0];
				}
				return true;
			}
		}

	return false;
}

bool ThreadEnumeration::fillBuffer() {
	bool worked = false;
	unique_lock<mutex> heapLock(mHeapLock, defer_lock_t());
	if (heapLock.try_lock()) {
		uint64_t split = mHeap[1].first;
		while ((mSplitBufferHead[split] != mSplitBufferTail[split])
				&& (mBufferHead + mBufferMask > mBufferTail)) {
			mBuffer[mBufferTail & mBufferMask] = mHeap[1].second;
			++mBufferTail;
			mHeap[1].second = mSplitBuffer[split][mSplitBufferHead[split]
					& mBufferMask];
			++mSplitBufferHead[split];
			maxHeap(1);
			split = mHeap[1].first;
			worked = true;
		}
	}
	return worked;
}

void ThreadEnumeration::pushHeap(uint64_t pSplit, const Candidate& pCandidate) {
	uint64_t index = mHeap.size();
	mHeap.emplace_back(pSplit, pCandidate);
	while (index > 1 && mHeap[index >> 1].second.first
			< mHeap[index].second.first) {
		swap(mHeap[index >> 1], mHeap[index]);
		index >>= 1;
	}
}

void ThreadEnumeration::maxHeap(uint64_t index) {
	uint64_t largest = index;
	uint64_t child = index << 1;
	if (child < mHeap.size() && mHeap[largest].second.first
			< mHeap[child].second.first)
		largest = child;
	++child;
	if (child < mHeap.size() && mHeap[largest].second.first
			< mHeap[child].second.first)
		largest = child;
	if (largest != index) {
		swap(mHeap[largest], mHeap[index]);
		maxHeap(largest);
	}
}
