/*
 * ThreadRank.cpp
 *
 *  Created on: Apr 12, 2012
 *      Author: nicolaq
 */

#include<iostream>

#include "ThreadRank.h"
#include "SplitDistribution.h"

ThreadRank::ThreadRank(const array<vector<double> , 0x10>& pDistribution,
		uint64_t pNumThreads, uint64_t pNumSplit) :
	mUnitSize(0x10000), mFound(false), mLastProbabilities(pNumSplit, 1.0), mRank(pNumSplit, 0),
			mNumThreads(pNumThreads), mEnumLocks(pNumSplit) {
auto splitMap = splitDistribution(indexDistribution(pDistribution), pNumSplit);
for(auto& v0 : splitMap)
mSplitEnums.push_back(unique_ptr<Enumeration<0x10>>(new Enumeration<0x10>(v0.second)));

}

ThreadRank::~ThreadRank() {
}

uint64_t ThreadRank::findRank(double pProbability) {
	mAttack = false;
	mTargetProbability = pProbability;
	mFound = false;
	for (unsigned i0 = 0; i0 < mNumThreads; ++i0)
	mThreads.push_back(thread {&ThreadRank::doWork, this});
	for(auto& v0: mThreads)
	v0.join();
	return accumulate(mRank.begin(), mRank.end(), 0);
}

Candidate ThreadRank::findKey(
		function<bool(array<unsigned char, 0x10> )>& pOracle) {
	mAttack = true;
	mFound = false;
	for (unsigned i0 = 0; i0 < mNumThreads; ++i0)
	mThreads.push_back(thread {&ThreadRank::doWork, this});
	for(auto& v0: mThreads)
	v0.join();
	return mCandidate;
}

uint64_t ThreadRank::memSize() const {
	uint64_t sizeValue = 0;
	for(auto& v0 : mSplitEnums)
	sizeValue += v0->memSize();
	sizeValue += sizeof(*this);
	sizeValue += mEnumLocks.size() * sizeof(mutex);
	sizeValue += mSplitEnums.size() * sizeof(Enumeration<0x10>*);
	return sizeValue;
}

uint64_t ThreadRank::diskSize() const {
	uint64_t sizeValue = 0;
	for(auto& v0 : mSplitEnums)
	sizeValue += v0->diskSize();
	return sizeValue;
}

void ThreadRank::doWork() {
	while (!mFound)
		doSomeWork();
}

bool ThreadRank::doSomeWork() {
	//*
	vector<uint64_t> index(mLastProbabilities.size());
	iota(index.begin(), index.end(), 0);
	sort(index.begin(), index.end(),
			[&](uint64_t v1, uint64_t v2) {return mLastProbabilities[v1] > mLastProbabilities[v2];});
	if (!mAttack && mLastProbabilities[index[0]] <= mTargetProbability)
		mFound = true;
	for (uint64_t i0 = 0; i0 < mSplitEnums.size(); i0++) {
		unique_lock<mutex> enumLock(mEnumLocks[index[i0]], defer_lock_t());
		if (enumLock.try_lock()) {
			double lastProbability = mLastProbabilities[index[i0]];
			for (uint64_t i1 = 0; (mAttack ? !mFound
					: (lastProbability >= mTargetProbability))
					&& i1 < mUnitSize; ++i1) {
				auto candidate = mSplitEnums[index[i0]]->nextCandidate();
				lastProbability = candidate.first;
				if (mAttack && mOracle(candidate.second)) {
					mCandidate = candidate;
					mFound = true;
				}
				++mRank[index[i0]];
			}
			mLastProbabilities[index[i0]] = lastProbability;
			return true;
		}
	}
	return false;
}

