/*
 * @file  keyEnumeration.cpp
 * @author  Nicolas Veyrat-Charvillon <nicolas.veyrat@uclouvain.be>
 * @date Jan 29, 2013
 * @version 1.1
 *
 * @section LICENSE
 *
 * Artistic License 2.0
 *
 * @ section DESCRIPTION
 *
 *  MATLAB hooks for key enumeration
 *
 * This is a MEX-file for MATLAB
 */

#include "mex.h"

#include "../Enumeration/Enumeration.h"

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray *prhs[]) {
	if(nrhs != 2)
		mexErrMsgTxt("Two arguments required: Event PMF and enumeration list size");
	if(nlhs != 1)
		mexErrMsgTxt("Too many output arguments");

	if(!(mxGetM(prhs[0]) == 0x100 && mxGetN(prhs[0]) == 0x10))
		mexErrMsgTxt("Event PMF must be 256 rows by 16 columns");
	if(!mxIsDouble(prhs[0]) || mxIsComplex(prhs[0]))
		mexErrMsgTxt("Event PMF must contain noncomplex scalar doubles");

	if(!(mxGetM(prhs[1]) == 1 && mxGetN(prhs[1]) == 1) || !mxIsDouble(prhs[1]))
		mexErrMsgTxt("Second argument must be an unsigned integer");

	double *pPmfInput = mxGetPr(prhs[0]);

	unsigned numEnum = mxGetScalar(prhs[1]);
	mwSize outD = 2;
	mwSize outMN[2];
	outMN[0] = numEnum;
	outMN[1] = 0x10;
	plhs[0] = mxCreateNumericArray(outD, outMN, mxUINT8_CLASS, mxREAL);

	array<vector<double>, 0x10> eventPMF;
	for(unsigned i0 = 0; i0 < 0x10; i0++)
		for(unsigned i1 = 0; i1 < 0x100; i1++)
			eventPMF[i0].push_back(pPmfInput[(i0 << 8) | i1]);

	Enumeration<0x10> enumeration(eventPMF);

	unsigned char *outputList = static_cast<unsigned char *>(mxGetData(plhs[0]));

	unsigned outputIndex = 0;
	for(unsigned i0 = 0; i0 < numEnum; i0++) {
		auto candidate = enumeration.nextCandidate();
		for(unsigned i1 = 0; i1 < 0x10; i1++) {
			outputList[i1 * numEnum + i0] = candidate.second[i1];
			outputIndex++;
		}
	}
}
