/*
 * KeyTope.cpp
 *
 *  Created on: May 14, 2012
 *      Author: nicolaq
 */

#include "KeyTope.h"

KeyTope::KeyTope(const vector<vector<double>>& pDistribution) :
		mCarved(false), mCarvedInside(false), mVolume(1) {
	for (auto& v0 : pDistribution) {
		mTopCorner.push_back(0);
		mBotCorner.push_back(v0.size() - 1);
	}
	mCarvePoint = mTopCorner;
	computeVolume();
}

KeyTope::KeyTope(const vector<unsigned>& pTop, const vector<unsigned>& pBot) :
		mTopCorner(pTop), mBotCorner(pBot), mCarved(false), mCarvedInside(
				false), mCarvePoint(pTop) {
	computeVolume();
}

KeyTope::~KeyTope() {
}

KeyTope KeyTope::splitLargest() {
	vector<double> ratios;
	for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0) {
		ratios.push_back(
				mCarvedInside ?
						(mBotCorner[i0] - mCarvePoint[i0]) :
						(mCarvePoint[i0] - mTopCorner[i0]));
		ratios.back() /= mBotCorner[i0] - mTopCorner[i0] + 1;
	}
	unsigned index = max_element(ratios.begin(), ratios.end()) - ratios.begin();
	auto topCorner = mTopCorner;
	auto botCorner = mBotCorner;
	if (mCarvedInside) {
		mBotCorner[index] = mCarvePoint[index];
		topCorner[index] = mCarvePoint[index] + 1;
	} else {
		mTopCorner[index] = mCarvePoint[index];
		botCorner[index] = mCarvePoint[index] - 1;
	}
	ratios[index] = 0;
	if (count_if(ratios.begin(), ratios.end(),
			[](double value) {return (value != 0);}) == 1) {
		// carved volume simplifies to a full volume
		index = max_element(ratios.begin(), ratios.end()) - ratios.begin();
		if (mCarvedInside)
			mTopCorner[index] = mCarvePoint[index] + 1;
		else
			mBotCorner[index] = mCarvePoint[index] - 1;
		mCarved = false;
		mCarvedVolume = 0;
		mCarvePoint = mTopCorner;
	}
	computeVolume();
	return KeyTope(topCorner, botCorner);
}

KeyTope KeyTope::splitSmallest() {
	vector<double> ratios;
	for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0) {
		ratios.push_back(
				mCarvedInside ?
						(mBotCorner[i0] - mCarvePoint[i0]) :
						(mCarvePoint[i0] - mTopCorner[i0]));
		ratios.back() /=
				mCarvedInside ?
						(mCarvePoint[i0] - mTopCorner[i0] + 1) :
						(mBotCorner[i0] - mCarvePoint[i0] + 1);
		if (ratios.back() == 0)
			ratios.back() = numeric_limits<double>::infinity();
	}
	unsigned index = min_element(ratios.begin(), ratios.end()) - ratios.begin();
	auto topCorner = mTopCorner;
	auto botCorner = mBotCorner;
	if (mCarvedInside) {
		topCorner[index] = mCarvePoint[index] + 1;
		mCarvePoint[index] = mBotCorner[index];
		botCorner = mCarvePoint;
	} else {
		botCorner[index] = mCarvePoint[index] - 1;
		mCarvePoint[index] = mTopCorner[index];
		topCorner = mCarvePoint;
	}
	ratios[index] = numeric_limits<double>::infinity();
	if (count_if(ratios.begin(), ratios.end(),
			[](double value) {return (value != numeric_limits<double>::infinity());})
			== 1) {
		// carved volume simplifies to a full volume
		index = min_element(ratios.begin(), ratios.end()) - ratios.begin();
		if (mCarvedInside)
			mTopCorner[index] = mCarvePoint[index] + 1;
		else
			mBotCorner[index] = mCarvePoint[index] - 1;
		mCarved = false;
		mCarvedVolume = 0;
		mCarvePoint = mTopCorner;
	}
	computeVolume();

	return KeyTope(topCorner, botCorner);
}

uint128 KeyTope::carvePiece(function<int(const vector<unsigned>&)> pFunc,
		const array<uint128, 3>& pEstimation) {
	const unsigned trialsCredit = 0x50;
	const unsigned restartCredit = 0x4;
	double bestLogInterval = -1;
	for (unsigned iSide = 0; iSide < 2; ++iSide) {
		bool fromTop = iSide;
		for (unsigned iRestart = restartCredit; iRestart-- > 0;) {
			auto trialCarvePoint = fromTop ? mTopCorner : mBotCorner;
			int cornerSign = pFunc(trialCarvePoint);
			auto nextCarvePoint = trialCarvePoint;
			unsigned index;
			int increment;
			increment = 2 * fromTop - 1;
			vector<unsigned> dimIncrement;
			for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
				dimIncrement.push_back(mBotCorner[i0] - mTopCorner[i0] + 1);
			// Move towards the surface
			bool improved = true;
			int trials = trialsCredit;
			while (trials >= 0) {
				index = rand() % mTopCorner.size();
				nextCarvePoint[index] += increment * dimIncrement[index];
				improved = nextCarvePoint[index] >= mTopCorner[index]
						&& nextCarvePoint[index] <= mBotCorner[index]
						&& cornerSign == pFunc(nextCarvePoint);
				if (!improved) {
					--trials;
					nextCarvePoint[index] -= increment * dimIncrement[index];
					dimIncrement[index] >>= 1;
					dimIncrement[index] |= 1;
				} else
					trials = trialsCredit;
			}
			improved = !equal(nextCarvePoint.begin(), nextCarvePoint.end(),
					trialCarvePoint.begin());
			trialCarvePoint = nextCarvePoint;

			uint128 trialVolume = 1;
			for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
				trialVolume *=
						fromTop ?
								(nextCarvePoint[i0] - mTopCorner[i0] + 1) :
								(mBotCorner[i0] - nextCarvePoint[i0] + 1);
/*			// Simulated Annealing
			double temperature = pow(2., 0);
			double coolingFactor = pow(2., -1);
			double coolingStop = pow(2., -20);
			uint128 volume = 0;
			while (improved && temperature > coolingStop) {
				temperature *= coolingFactor;
				int steps = computeCredit;
				bool improved = false;
				while (--steps) {
					// take a valid step...
					bool valid = false;

					while (!valid) {
						increment = 2 * (rand() & 0x1) - 1;
						index = rand() % mTopCorner.size();
						nextCarvePoint[index] += increment;
						valid = nextCarvePoint[index] >= mTopCorner[index]
								&& nextCarvePoint[index] <= mBotCorner[index]
								&& cornerSign == pFunc(nextCarvePoint);
						if (!valid)
							nextCarvePoint[index] -= increment;
					}
					uint128 vol = 1;
					for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
						vol *= fromTop ?
								(nextCarvePoint[i0] - mTopCorner[i0] + 1) :
								(mBotCorner[i0] - nextCarvePoint[i0] + 1);
					if (vol > volume
							|| exp(
									(log(double(vol)) - log(double(volume)))
											/ temperature)
									> (rand() / double(RAND_MAX)))
						volume = vol;
					else
						nextCarvePoint[index] -= increment;
					if (volume > trialVolume) {
						improved = true;
						trialVolume = volume;
						steps = computeCredit;
					}
				}

				if (improved) {
					temperature /= coolingFactor;
					trialCarvePoint = nextCarvePoint;
				} else
					nextCarvePoint = trialCarvePoint;
			}*/

			double trialLogInterval = 0;
			if (fromTop) {
				trialLogInterval = log(double(pEstimation[0]));
				if(trialLogInterval < 0)
					trialLogInterval = 128;
				trialLogInterval -= log(double(pEstimation[2] + trialVolume));
			}
			else
				trialLogInterval = log(double(pEstimation[0] - trialVolume))
						- log(double(pEstimation[2]));
			if (bestLogInterval < 0 || trialLogInterval < bestLogInterval) {
				bestLogInterval = trialLogInterval;
				mCarved = true;
				mCarvedSign = cornerSign;
				mCarvedInside = fromTop;
				mCarvePoint = trialCarvePoint;
				iRestart = restartCredit;
			}
		}
	}
	computeVolume();
	auto carvedVolume = mCarvedVolume;
	unsigned numDiff = 0;
	for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
		numDiff += mCarvePoint[i0]
				!= (mCarvedInside ? mBotCorner[i0] : mTopCorner[i0]);
	if (numDiff == 1) {
		unsigned index = 0;
		for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
			if (mCarvePoint[i0]
					!= (mCarvedInside ? mBotCorner[i0] : mTopCorner[i0]))
				index = i0;
		mCarved = false;
		if (mCarvedInside)
			mTopCorner[index] = mCarvePoint[index] + 1;
		else
			mBotCorner[index] = mCarvePoint[index] - 1;
		mCarvePoint = mTopCorner;
		computeVolume();
	}
	return carvedVolume;
}

bool KeyTope::isCarved() const {
	return mCarved;
}

int KeyTope::carvedSign() const {
	return mCarvedSign;
}

uint128 KeyTope::volume() const {
	return mVolume;
}

bool KeyTope::operator()(const KeyTope& pTope1, const KeyTope& pTope2) const {
	return pTope1.volume() > pTope2.volume();
}

void KeyTope::computeVolume() {
	mVolume = 1;
	for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
		mVolume *= mBotCorner[i0] - mTopCorner[i0] + 1;
	if (mCarved) {
		mCarvedVolume = 1;
		if (mCarvedInside)
			for (unsigned i0 = 0; i0 < mTopCorner.size(); ++i0)
				mCarvedVolume *= mCarvePoint[i0] - mTopCorner[i0] + 1;
		else
			for (unsigned i0 = 0; i0 < mBotCorner.size(); ++i0)
				mCarvedVolume *= mBotCorner[i0] - mCarvePoint[i0] + 1;
		mVolume -= mCarvedVolume;
	}
}
