/*
 * RankKey.h
 *
 *  Created on: May 14, 2012
 *      Author: nicolaq
 */

#ifndef RANKKEY_H_
#define RANKKEY_H_

#include"Includes.h"
#include"TopeQueue.h"
#include"KeyTope.h"

/*
 * The class used to perform rank estimations
 */
class RankKey {
public:
	// Takes a list of subkey probabilities sets, and the probability of the correct key
	RankKey(const vector<vector<double>>& pDistribution,
			double pKeyProbability);
	virtual ~RankKey();

	// Merges subkey lists up to sizes 2^logSize
	// this improves convergence speed at the cost of memory usage
	// 24 is usually the best choice for a desktop
	void mergeDistribution(unsigned logSize = 24);
	// runs an estimation for the given amount of time, and outputs:
	// log2(lower bound), log2(number of equal candidates), log2(upper bound)
	// the second value may be -inf when the convergence hasn't found any exact match yet
	array<uint128, 3> operator()(unsigned pSeconds);
protected:
	void sortDistribution();
	double probability(const vector<unsigned>& pPoint) const;

	vector<vector<double>> mDistribution;
	double mKeyProbability;

	array<uint128, 3> mEstimation;
	TopeQueue mTopes;
};

#endif /* RANKKEY_H_ */
